;;;; AVL tree for integer sets
;
; Node = #(HEIGHT START END LEFT RIGHT)
; The empty node is #f

(import (chicken fixnum))

(define (t-start t) (vector-ref t 1))
(define (t-end t) (vector-ref t 2))
(define (t-left t) (vector-ref t 3))
(define (t-right t) (vector-ref t 4))

(define (t-height t) (if t (vector-ref t 0) 0))

(define (t-lookup i t)
  (and t
       (cond ((fx< i (t-start t)) (t-lookup i (t-left t)))
             ((fx> i (t-end t)) (t-lookup i (t-right t)))
             (else #t))))

(define (t-insert s e t)
  (cond ((not t) (t-root s e))
        ((fx< e (t-start t))
         (t-node (t-start t)
                 (t-end t)
                 (t-insert s e (t-left t))
                 (t-right t)))
        ((fx> s (t-end t))
         (t-node (t-start t)
                 (t-end t)
                 (t-left t)
                 (t-insert s e (t-right t))))
        (else (error 't-insert "range overlap" s e))))

(define (t-root s e) (t-node s e #f #f))

(define (t-node s e l r)
  (let ((hl (t-height l))
        (hr (t-height r)))
    (t-node2 (fx- hr hl) (fx+ 1 (fxmax hl hr)) s e l r)))

(define (t-node2 b h s e l r)
  (cond ((eq? -2 b) (t-node-il (t-balance l) h s e l r))
        ((eq? 2 b) (t-node-ir (t-balance r) h s e l r))
        (else (vector h s e l r))))

(define (t-balance t)
  (if t
      (fx- (t-height (t-right t)) (t-height (t-left t)))
      0))

(define (t-node-il n h s e l r)
  (cond ((fx< n 0) (t-rotate-r (vector h s e l r)))
        ((fx> n 0) (t-node s e (t-rotate-l l) r))
        (else (vector h s e l r))))

(define (t-node-ir n h s e l r)
  (cond ((fx< n 0) (t-node s e l (t-rotate-r r)))
        ((fx> n 0) (t-rotate-l (vector h s e l r)))
        (else (vector h s e l r))))

(define (t-rotate-l t)
  (let ((r (t-right t)))
    (t-node (t-start r)
            (t-end r)
            (t-node (t-start t) (t-end t) (t-left t) (t-left r))
            (t-right r))))

(define (t-rotate-r t)
  (let ((l (t-left t)))
    (t-node (t-start l)
            (t-end l)
            (t-left l)
            (t-node (t-start t) (t-end t) (t-right l) (t-right t)))))

(define (t-describe t)
  (let desc ((t t))
    (when t
      (desc (t-left t))
      (print (t-start t) " - " (t-end t))
      (desc (t-right t)))))

(define (t-foldl t x p) (t-foldl* t x p (lambda (y) y)))

(define (t-foldl* t x p k)
  (if t
      (t-foldl* (t-left t) x p
                (lambda (y)
                  (t-foldl* (t-right t)
                            (p y (t-start t) (t-end t))
                            p k)))
      (k x)))

(define (t->list t)
  (reverse (t-foldl t '() (lambda (x s e) (cons (cons s e) x)))))

(define (list->t lst)
  (let loop ((t #f)
             (lst lst))
    (if (null? lst)
        t
        (let ((s (caar lst))
              (e (cdar lst)))
          (cond ((fx< e s) (loop t (cdr lst)))
                ((and (pair? (cdr lst))
                      (fx>= e (fx- (caadr lst) 1)))
                 (loop t (cons (cons s (cdadr lst)) (cddr lst))))
                (else
                  (loop (t-insert s e t) (cdr lst))))))))

(define (t-union t1 t2)
  (list->t
    (let loop ((tl1 (t->list t1))
               (tl2 (t->list t2)))
      (cond ((null? tl1) tl2)
            ((null? tl2) tl1)
            (else
              (let ((s1 (caar tl1))
                    (e1 (cdar tl1))
                    (s2 (caar tl2))
                    (e2 (cdar tl2)))
                (cond ((fx> s2 (fx+ e1 1))
                	   ;; interval 2 is beyond interval 1
                       (cons (car tl1) (loop (cdr tl1) tl2)))
                      ((fx< (fx+ e2 1) s1)
                       ;; interval 2 is before interval 1
                       (cons (car tl2) (loop tl1 (cdr tl2))))
                      (else
                        ;; intervals overlap
                        (let ((m (cons (fxmin s1 s2)
                        			   (fxmax e1 e2))))
					      (if (fx> e1 e2)
					          (loop (cons m (cdr tl1))
					          		(cdr tl2))
					          (loop (cdr tl1)
					          	    (cons m (cdr tl2)))))))))))))

(define (t-intersection t1 t2)
  (list->t
    (let loop ((tl1 (t->list t1))
               (tl2 (t->list t2)))
      (cond ((null? tl1) '())
            ((null? tl2) '())
            (else
              (let ((s1 (caar tl1))
                    (e1 (cdar tl1))
                    (s2 (caar tl2))
                    (e2 (cdar tl2)))
                (cond ((fx> s2 e1) (loop (cdr tl1) tl2))
                      ((fx< e2 s1) (loop tl1 (cdr tl2)))
                      ((fx> e2 e1)
                       (cons (cons (fxmax s1 s2) e1)
                             (loop (cdr tl1)
                                   (cons (cons (fx+ e1 1) e2)
                                         tl2))))
                      ((fx> e1 e2)
                       (cons (cons (fxmax s1 s2) e2)
                             (loop (cons (cons (fx+ e2 1) e1) tl1)
                                   (cdr tl2))))
                      (else
                        (cons (cons (fxmax s1 s2) (fxmin e1 e2))
                              (loop (cdr tl1) (cdr tl2)))))))))))

(define (t-xor t1 t2)
  (list->t
    (let loop ((tl1 (t->list t1))
               (tl2 (t->list t2)))
      (cond ((null? tl1) tl2)
            ((null? tl2) tl1)
            (else
              (let ((s1 (caar tl1))
                    (e1 (cdar tl1))
                    (s2 (caar tl2))
                    (e2 (cdar tl2)))
                (cond ((fx> s2 e1)
                       (cons (car tl1) (loop (cdr tl1) tl2)))
                      ((fx< e2 s1)
                       (cons (car tl2) (loop tl1 (cdr tl2))))
                      ((fx> s2 s1)
                       (cons (cons s1 (fx- s2 1))
                             (loop (cons (cons s2 e1) (cdr tl1))
                                   (cons (cons s2 e2) (cdr tl2)))))
                      ((fx> s1 s2)
                       (cons (cons s2 (fx- s1 1))
                             (loop (cons (cons s1 e1) (cdr tl1))
                                   (cons (cons s1 e2) (cdr tl2)))))
                      ((fx> e1 e2)
                       (loop (cons (cons (fx+ e2 1) e1) (cdr tl1))
                             (cdr tl2)))
                      (else
                        (cons (cons (fx+ e1 1) e2)
                              (loop (cdr tl1) (cdr tl2)))))))))))

(define (t-difference t1 t2)
  (list->t
    (let loop ((tl1 (t->list t1))
               (tl2 (t->list t2)))
      (cond ((null? tl1) '())
            ((null? tl2) tl1)
            (else
              (let ((s1 (caar tl1))
                    (e1 (cdar tl1))
                    (s2 (caar tl2))
                    (e2 (cdar tl2)))
                (cond ((fx> s2 e1)
                       (cons (car tl1) (loop (cdr tl1) tl2)))
                      ((fx< e2 s1) (loop tl1 (cdr tl2)))
                      ((fx> s2 s1)
                       (cons (cons s1 (fx- s2 1))
                             (loop (cons (cons s2 e1) (cdr tl1))
                                   (cons (cons s2 e2) (cdr tl2)))))
                      ((fx> e1 e2)
                       (loop (cons (cons (fx+ e2 1) e1) (cdr tl1))
                             (cdr tl2)))
                      (else (loop (cdr tl1) (cdr tl2))))))))))

(define (t-complement t from to)
  (list->t
    (let loop ((tl (t->list t))
               (from from)
               (to to))
      (cond ((fx> from to) '())
            ((null? tl) (list (cons from to)))
            (else
              (let ((s (caar tl))
                    (e (cdar tl)))
                (cond ((fx> s from)
                       (cons (cons from (fx- s 1))
                             (loop (cdr tl)
                                   (fx+ e 1)
                                   to)))
                      (else (loop (cdr tl)
                                  (fx+ e 1)
                                  to)))))))))

(define (t-subset? t1 t2)
  (let loop ((tl1 (t->list t1))
             (tl2 (t->list t2)))
    (cond ((null? tl1))
          ((null? tl2) #f)
          ((fx< (cdar tl2) (caar tl1)) (loop tl1 (cdr tl2)))
          ((or (fx< (caar tl1) (caar tl2))
               (fx> (cdar tl1) (cdar tl2)))
           #f)
         (else (loop (cdr tl1) tl2)))))
