;;;; srfi-14.scm - Shivers' reference implementation of SRFI-14

(declare
  (fixnum)
  (disable-interrupts))

(module srfi-14
  (->char-set
   char-set
   char-set->list
   char-set->string
   char-set-adjoin
   char-set-adjoin!
   char-set-any
   char-set-complement
   char-set-complement!
   char-set-contains?
   char-set-copy
   char-set-count
   char-set-cursor
   char-set-cursor-next
   char-set-delete
   char-set-delete!
   char-set-diff+intersection
   char-set-diff+intersection!
   char-set-difference
   char-set-difference!
   char-set-every
   char-set-filter
   char-set-filter!
   char-set-fold
   char-set-for-each
   char-set-hash
   char-set-intersection
   char-set-intersection!
   char-set-map
   char-set-ref
   char-set-size
   char-set-unfold
   char-set-unfold!
   char-set-union
   char-set-union!
   char-set-xor
   char-set-xor!
   char-set:ascii
   char-set:blank
   char-set:digit
   char-set:empty
   char-set:full
   char-set:graphic
   char-set:hex-digit
   char-set:iso-control
   char-set:letter
   char-set:letter+digit
   char-set:lower-case
   char-set:printing
   char-set:punctuation
   char-set:symbol
   char-set:title-case
   char-set:upper-case
   char-set:whitespace
   char-set<=
   char-set=
   char-set?
   end-of-char-set?
   list->char-set
   list->char-set!
   string->char-set
   string->char-set!
   ucs-range->char-set
   ucs-range->char-set!)

(import (scheme)
        (chicken base)
        (chicken sort)
        srfi-1
        (chicken fixnum)
        (chicken platform))

(register-feature! 'srfi-14)

(define min-ucs 0)
(define max-ucs #x10ffff)

(include "iavl.scm")

;;; SRFI-14 character-sets library                              -*- Scheme -*-
;;;
;;; - Ported from MIT Scheme runtime by Brian D. Carlstrom.
;;; - Massively rehacked & extended by Olin Shivers 6/98.
;;; - Massively redesigned and rehacked 5/2000 during SRFI process.
;;; At this point, the code bears the following relationship to the
;;; MIT Scheme code: "This is my grandfather's axe. My father replaced
;;; the head, and I have replaced the handle." Nonetheless, we preserve
;;; the MIT Scheme copyright:
;;;     Copyright (c) 1988-1995 Massachusetts Institute of Technology
;;; The MIT Scheme license is a "free software" license. See the end of
;;; this file for the tedious details. 

;;; Exports:
;;; char-set? char-set= char-set<=
;;; char-set-hash 
;;; char-set-cursor char-set-ref char-set-cursor-next end-of-char-set?
;;; char-set-fold char-set-unfold char-set-unfold!
;;; char-set-for-each char-set-map
;;; char-set-copy
;;;
;;; char-set  list->char-set  string->char-set 
;;; char-set! list->char-set! string->char-set! 
;;;
;;; filterchar-set  ucs-range->char-set  ->char-set
;;; filterchar-set! ucs-range->char-set!
;;;
;;; char-set->list char-set->string
;;;
;;; char-set-size char-set-count char-set-contains?
;;; char-set-every char-set-any
;;;
;;; char-set-adjoin  char-set-delete 
;;; char-set-adjoin! char-set-delete!
;;; 
;;; char-set-complement  char-set-union  char-set-intersection  char-set-difference
;;; char-set-complement! char-set-union! char-set-intersection! char-set-difference!
;;;
;;; char-set-difference char-set-xor  char-set-diff+intersection
;;; char-set-difference! char-set-xor! char-set-diff+intersection!
;;;
;;; char-set:lower-case         char-set:upper-case     char-set:title-case
;;; char-set:letter             char-set:digit          char-set:letter+digit
;;; char-set:graphic            char-set:printing       char-set:whitespace
;;; char-set:iso-control        char-set:punctuation    char-set:symbol
;;; char-set:hex-digit          char-set:blank          char-set:ascii
;;; char-set:empty              char-set:full
;;;
;;; This is simple code, not great code.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-char-set s) (##sys#make-structure 'char-set s))
(define (char-set:s cs) (force (##sys#slot cs 1)))
(define (char-set:set! cs t) (##sys#setslot cs 1 t))
(define (char-set? x) (##sys#structure? x 'char-set))

;;; Parse, type-check & default a final optional BASE-CS parameter from
;;; a rest argument. Return a *fresh copy* of the underlying string.
;;; The default is the empty set. The PROC argument is to help us
;;; generate informative error exceptions.

(define (%default-base maybe-base proc)
  (if (pair? maybe-base)
      (let ((bcs  (car maybe-base))
            (tail (cdr maybe-base)))
        (if (null? tail)
            (if (char-set? bcs) 
                (char-set:s bcs)
                (##sys#error proc "BASE-CS parameter not a char-set" bcs))
            (##sys#error proc "Expected final base char set -- too many parameters"
                         maybe-base)))
      #f))

;;; If CS is really a char-set, do CHAR-SET:S, otw report an error msg on
;;; behalf of our caller, PROC. This procedure exists basically to provide
;;; explicit error-checking & reporting.

(define (%char-set:s/check cs proc)
  (if (char-set? cs) 
      (char-set:s cs)
      (##sys#error proc "Not a char-set" cs)))

;;; These internal functions hide a lot of the dependency on the
;;; underlying string representation of char sets. They should be
;;; inlined if possible.

(define (char-set-copy cs)
  (make-char-set (%char-set:s/check cs 'char-set-copy)))

(define char-set= 
  (lambda rest
    (or (null? rest)
        (let* ((cs1  (car rest))
               (rest (cdr rest))
               (s1 (%char-set:s/check cs1 'char-set=)))
          (let lp ((rest rest))
            (or (not (pair? rest))
                (and (equal? s1 (%char-set:s/check (car rest) 'char-set=))
                     (lp (cdr rest)))))))))

(define char-set<=
  (lambda rest
    (or (null? rest)
        (let ((cs1  (car rest))
              (rest (cdr rest)))
          (let lp ((s1 (%char-set:s/check cs1 'char-set<=))
                   (rest rest))
            (or (not (pair? rest))
                (let ((s2 (%char-set:s/check (car rest) 'char-set<=))
                      (rest (cdr rest)))
                  (if (eq? s1 s2)
                      (lp s2 rest) ; Fast path
                      (and (t-subset? s1 s2)
                           (lp s2 rest))))))))))

;;; Hash
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Compute (c + 37 c + 37^2 c + ...) modulo BOUND.
;;; If you keep BOUND small enough, the intermediate calculations will 
;;; always be fixnums. How small is dependent on the underlying Scheme system; 
;;; we use a default BOUND of 2^22 = 4194304, which should hack it in
;;; Schemes that give you at least 29 signed bits for fixnums. The core 
;;; calculation that you don't want to overflow is, worst case,
;;;     (+ 65535 (* 37 (- bound 1)))
;;; where 65535 is the max character code. Choose the default BOUND to be the
;;; biggest power of two that won't cause this expression to fixnum overflow, 
;;; and everything will be copacetic.

(define (char-set-hash cs . maybe-bound)
  (let ((bound (optional maybe-bound 4194304)))
    (##sys#check-fixnum bound 'char-set-hash)
    (if (zero? bound) (set! bound 4194304))
    (let* ((s (%char-set:s/check cs 'char-set-hash))
           ;; Compute a 111...1 mask that will cover BOUND-1:
           (mask (let lp ((i #x10000))  ; Let's skip first 16 iterations, eh?
                   (if (>= i bound) (- i 1) (lp (+ i i))))))
      (fxmod (t-foldl s 0
                      (lambda (x st en)
                        (do ((i st (fx+ i 1))
                             (r x (fxand mask (+ (* 37 r) i))))
                            ((fx> i en) r))))
             bound))))

(define (char-set-contains? cs char)
  (##sys#check-char char 'char-set-contains?)
  (t-lookup (char->integer char)
            (%char-set:s/check cs 'char-set-contains?)))

(define (char-set-size cs)
  (let ((s (%char-set:s/check cs 'char-set-size)))
    (t-foldl s 0 (lambda (x s e) (fx+ 1 (fx+ (fx- e s) x))))))

(define (char-set-count pred cset)
  (let ((s (%char-set:s/check cset 'char-set-count)))
    (t-foldl s 0 (lambda (x s e)
                   (do ((i s (fx+ i 1))
                        (n 0))
                       ((fx> i e) (fx+ x n))
                       (when (pred (integer->char i))
                         (set! n (fx+ n 1))))))))

;;; -- Adjoin & delete

(define (ucs->t ucs)
  (let loop ((cs (sort ucs fx<=))
             (new '())
             (from #f)
             (to #f))
    (cond ((null? cs)
           (list->t (reverse (if from (cons (cons from to) new) new))))
          ((and from (eq? (car cs) (fx+ to 1)))
           (loop (cdr cs) new from (car cs)))
          (from (loop (cdr cs) (cons (cons from to) new) (car cs) (car cs)))
          (else (loop (cdr cs) new (car cs) (car cs))))))

(define (char-set-adjoin cs . chars)
  (make-char-set
    (t-union (%char-set:s/check cs 'char-set-adjoin)
             (ucs->t (map (lambda (c)
                            (##sys#check-char c 'char-set-adjoin)
                            (char->integer c))
                       chars)))))
                       
(define (char-set-adjoin! cs . chars)
  (char-set:set! cs (char-set:s (apply char-set-adjoin cs chars)))
  cs)

(define (char-set-delete cs . chars)
  (make-char-set
    (t-difference (%char-set:s/check cs 'char-set-delete)
                  (ucs->t (map (lambda (c)
                            (##sys#check-char c 'char-set-delete)
                            (char->integer c))
                       chars)))))

(define (char-set-delete! cs . chars)
  (char-set:set! cs (char-set:s (apply char-set-delete cs chars)))
  cs)

;;; Cursors

(define-record char-set-cursor list)

(define (char-set-cursor cset)
  (let ((s (%char-set:s/check cset 'char-set-cursor)))
    (make-char-set-cursor (t->list s))))
  
(define (end-of-char-set? cursor)
  (null? (char-set-cursor-list cursor)))

(define (char-set-ref cset cursor)
  (let ((lst (char-set-cursor-list cursor)))
    (if (null? lst)
        (##sys#error 'char-set-ref "char-set cursor beyond limit" cursor cset)
        (integer->char (caar lst)))))

(define (char-set-cursor-next cset cursor)
  (let ((lst (char-set-cursor-list cursor)))
    (cond ((null? lst)
           (##sys#error 'char-set-cursor-next "char-set cursor beyond limit"
                        cursor cset))
          ((eq? (caar lst) (cdar lst))
           (make-char-set-cursor (cdr lst)))
          (else (make-char-set-cursor (cons (cons (fx+ (caar lst) 1) (cdar lst))
                                            (cdr lst)))))))

;;; -- for-each map fold unfold every any

(define (char-set-for-each proc cs)
  (let ((s (%char-set:s/check cs 'char-set-for-each)))
    (t-foldl s #f (lambda (_ s e)
                    (do ((i s (fx+ i 1)))
                        ((fx> i e))
                        (proc (integer->char i)))))))

(define (char-set-map proc cs)
  (let* ((s (%char-set:s/check cs 'char-set-map))
         (lst (t->list s)))
    (make-char-set
      (ucs->t
        (append-map
          (lambda (rng)
            (let ((e (cdr rng)))
              (let loop ((i (car rng)))
                (if (fx> i e)
                    '()
                    (cons (char->integer (proc (integer->char i)))
                          (loop (fx+ i 1)))))))
          lst)))))

(define (char-set-fold kons knil cs)
  (let* ((s (%char-set:s/check cs 'char-set-fold))
         (lst (t->list s)))
    (let outer ((lst lst)
                (ans knil))
      (if (null? lst)
          ans
          (let ((e (cdar lst)))
            (let inner ((i (caar lst))
                        (ans ans))
              (if (fx> i e)
                  (outer (cdr lst) ans)
                  (inner (fx+ i 1)
                       (kons (integer->char i) ans)))))))))

(define (char-set-every pred cs)
  (let* ((s (%char-set:s/check cs 'char-set-every))
         (lst (t->list s)))
    (let outer ((lst lst))
      (or (null? lst)
          (let ((e (cdar lst)))
            (let inner ((i (caar lst)))
              (if (fx> i e)
                  (outer (cdr lst))
                  (and (pred (integer->char i))
                       (inner (fx+ i 1))))))))))

(define (char-set-any pred cs)
  (let* ((s (%char-set:s/check cs 'char-set-any))
         (lst (t->list s)))
    (let outer ((lst lst))
      (and (pair? lst)
           (let ((e (cdar lst)))
             (let inner ((i (caar lst)))
               (if (fx> i e)
                   (outer (cdr lst))
                   (or (pred (integer->char i))
                       (inner (fx+ i 1))))))))))

(define (%char-set-unfold! proc p f g s seed)
  (let lp ((seed seed))
    (cond ((not (p seed))                       ; P says we are done.
           (let ((x (char->integer (f seed))))  ; Add (F SEED) to set.
             (set! s (t-union (list->t (list (cons x x))) s))
             (lp (g seed))))                    ; Loop on (G SEED).
          (else s))))

(define (char-set-unfold p f g seed . maybe-base)
  (let ((bs (%default-base maybe-base char-set-unfold)))
    (make-char-set (%char-set-unfold! char-set-unfold p f g bs seed))))

(define (char-set-unfold! p f g seed base-cset)
  (char-set:set! base-cset 
                 (%char-set-unfold! char-set-unfold! p f g
                                    (%char-set:s/check base-cset 'char-set-unfold!)
                                    seed))
  base-cset)

;;; list <--> char-set

(define (%list->char-set! chars s)
  (t-union s (ucs->t (map char->integer chars))))

(define (char-set . chars)
  (make-char-set (%list->char-set! chars #f)))

(define (list->char-set chars . maybe-base)
  (let ((bs (%default-base maybe-base list->char-set)))
    (make-char-set (%list->char-set! chars bs))))

(define (list->char-set! chars base-cs)
  (char-set:set! base-cs
                 (%list->char-set! chars 
                                   (%char-set:s/check base-cs 'list->char-set!)))
  base-cs)

(define (char-set->list cs)
  (let ((s (%char-set:s/check cs 'char-set->list)))
    (append-map
      (lambda (rng)
        (let ((s (car rng)))
          (do ((i (cdr rng) (fx- i 1))
               (lst '() (cons (integer->char i) lst)))
              ((fx< i s) lst))))
      (t->list s))))

;;; string <--> char-set

(define (%string->char-set! str s proc)
  (##sys#check-string str proc)
  (t-union s (ucs->t (map char->integer (string->list str)))))

(define (string->char-set str . maybe-base)
  (let ((bs (%default-base maybe-base 'string->char-set)))
    (make-char-set (%string->char-set! str bs 'string->char-set))))

(define (string->char-set! str base-cs)
  (let ((s (%char-set:s/check base-cs 'string->char-set!)))
    (char-set:set! base-cs (%string->char-set! str s 'string->char-set!))
    base-cs))

(define (char-set->string cs)
  (let ((s (%char-set:s/check cs 'char-set->string)))
    (list->string
      (reverse
        (let loop ((lst (t->list s)))
          (if (null? lst)
              '()
              (let ((e (cdar lst)))
                (do ((i (caar lst) (fx+ i 1))
                     (l2 '() (cons (integer->char i) l2)))
                    ((fx> i e) 
                     (append l2 (loop (cdr lst))))))))))))

;;; -- UCS-range -> char-set

(define (%ucs-range->char-set! lower upper error? bs proc)
  (##sys#check-fixnum lower proc)
  (##sys#check-fixnum upper proc)
  (when (and error (or (fx< lower min-ucs) (fx> upper max-ucs)))
    (##sys#error proc "invalid UCS range" lower upper))
  (t-union bs (t-root (fxmax min-ucs lower) (fxmin max-ucs (fx- upper 1)))))

(define (ucs-range->char-set lower upper . rest)
  (let-optionals* rest ((error? #f) rest)
    (let ((bs (%default-base rest 'ucs-range->char-set)))
      (make-char-set
        (%ucs-range->char-set! lower upper error? bs 'ucs-range->char-set)))))

(define (ucs-range->char-set! lower upper error? base-cs)
  (let ((s (%char-set:s/check base-cs 'ucs-range->char-set!)))
    (char-set:set! base-cs (%ucs-range->char-set! lower upper error? s
                                                  'ucs-range->char-set!))
    base-cs))

;;; -- predicate -> char-set

(define (%char-set-filter! pred bs)
  (list->t
    (let loop ((lst (t->list bs)))
      (if (null? lst)
          '()
          (let ((s (caar lst))
                (e (cdar lst)))
            (let inner ((i s)
                        (p #f))
              (cond ((fx> i e)
                     (if p
                         (cons (cons p e) (loop (cdr lst)))
                         (loop (cdr lst))))
                    ((pred (integer->char i))
                     (inner (fx+ i 1) (or p i)))
                    (p (cons (cons p (fx- i 1)) (inner (fx+ i 1) #f)))
                    (else (inner (fx+ i 1) #f)))))))))

(define (char-set-filter predicate domain . maybe-base)
  (let ((bs (%default-base maybe-base 'char-set-filter)))
    (make-char-set
      (t-union bs
               (%char-set-filter! predicate
                                  (%char-set:s/check domain 'char-set-filter!))))))

(define (char-set-filter! predicate domain base-cs)
  (let ((s (%char-set:s/check base-cs 'char-set-filter!)))
    (char-set:set! base-cs
                   (t-union s (%char-set-filter! predicate
                                                 (%char-set:s/check domain 'char-set-filter!))))
    base-cs))

;;; {string, char, char-set, char predicate} -> char-set

(define (->char-set x)
  (cond ((char-set? x) x)
        ((string? x) (string->char-set x))
        ((char? x) (char-set x))
        (else (##sys#error '->char-set "Not a charset, string or char." x))))

;;; Set algebra
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The exported ! procs are "linear update" -- allowed, but not required, to
;;; side-effect their first argument when computing their result. In other
;;; words, you must use them as if they were completely functional, just like
;;; their non-! counterparts, and you must additionally ensure that their
;;; first arguments are "dead" at the point of call. In return, we promise a
;;; more efficient result, plus allowing you to always assume char-sets are
;;; unchangeable values.

;;; -- Invert

(define (char-set-complement cs)
  (let ((s (%char-set:s/check cs 'char-set-complement)))
    (make-char-set (t-complement s min-ucs max-ucs))))

(define (char-set-complement! cset)
  (let ((s (%char-set:s/check cset 'char-set-complement!)))
    (char-set:set! cset (t-complement s min-ucs max-ucs))
    cset))

;;; -- Union

(define (char-set-union! cset1 . csets)
  (let ((s (%char-set:s/check cset1 'char-set-union!)))
    (char-set:set! cset1
                   (let loop ((cs csets)
                              (s s))
                     (if (null? cs)
                         s
                         (let ((s2 (%char-set:s/check (car cs) 'char-set-union!)))
                           (loop (cdr cs)
                                 (t-union s s2))))))
    cset1))

(define (char-set-union . csets)
  (apply char-set-union! (make-char-set #f) csets))

;;; -- Intersection

(define (char-set-intersection! cset1 . csets)
  (let ((s (%char-set:s/check cset1 'char-set-intersection!)))
    (char-set:set! cset1 
                   (let loop ((cs csets)
                              (s s))
                     (if (null? cs)
                         s
                         (let ((s2 (%char-set:s/check (car cs) 'char-set-intersection!)))
                           (loop (cdr cs)
                                 (t-intersection s s2))))))
    cset1))

(define (char-set-intersection . csets)
  (if (pair? csets)
      (apply char-set-intersection! (make-char-set (char-set:s (car csets)))
             (cdr csets))
      (make-char-set (list->t (list (cons min-ucs max-ucs))))))

;;; -- Difference

(define (char-set-difference! cset1 . csets)
  (let ((s (%char-set:s/check cset1 'char-set-difference!)))
    (char-set:set! cset1 
                   (let loop ((cs csets)
                              (s s))
                     (if (null? cs)
                         s
                         (let ((s2 (%char-set:s/check (car cs) 'char-set-difference!)))
                           (loop (cdr cs)
                                 (t-difference s s2))))))
    cset1))

(define (char-set-difference cs1 . csets)
  (let ((s2 (make-char-set (char-set:s cs1))))
    (if (pair? csets)
        (apply char-set-difference! s2 csets)
        s2)))

;;; -- Xor

(define (char-set-xor! cset1 . csets)
  (let ((s (%char-set:s/check cset1 'char-set-xor!)))
    (char-set:set! cset1 
                   (let loop ((cs csets)
                              (s s))
                     (if (null? cs)
                         s
                         (let ((s2 (%char-set:s/check (car cs) 'char-set-difference!)))
                           (loop (cdr cs)
                                 (t-xor s s2))))))
    cset1))

(define (char-set-xor . csets)
  (let ((s2 (make-char-set #f)))
    (if (pair? csets)
        (apply char-set-xor! s2 csets)
        s2)))

;;; -- Difference & intersection

(define (char-set-diff+intersection! cs1 cs2 . csets)
  (let ((s1 (apply char-set-difference cs1 cs2 csets))
        (s2 (apply char-set-intersection cs1 cs2 csets)))
    (char-set:set! cs1 (char-set:s s1))
    (char-set:set! cs2 (char-set:s s2))
    (values cs1 cs2)))

(define (char-set-diff+intersection cs1 . csets)
  (let ((s1 (apply char-set-difference cs1 csets))
        (s2 (apply char-set-intersection cs1 csets)))
    (values s1 s2)))

;;;; System character sets
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; If your Scheme implementation allows you to mark the underlying strings
;;; as immutable, you should do so -- it would be very, very bad if a client's
;;; buggy code corrupted these constants.

(define (decompress data #!optional base)
  (let ((s base)
        (len (vector-length data)))
    (do ((i 0 (fx+ i 1)))
        ((fx>= i len) s)
        (let ((x (vector-ref data i)))
          (if (pair? x)
              (set! s (t-insert (car x) (cdr x) s))
              (set! s (t-insert x x s)))))))

(define-syntax define-char-set 
  (syntax-rules ()
    ((_ name body)
     (define name (make-char-set (delay body))))))

(define char-set:empty (char-set))
(define char-set:full (ucs-range->char-set min-ucs max-ucs))

(define-char-set char-set:lower-case
  (decompress '#((#x0061 . #x007A) (#x00DF . #x00F6) (#x00F8 . #x00FF) (#x0137 . #x0138) (#x0148 . #x0149) (#x017E . #x0180) (#x018C . #x018D) (#x0199 . #x019B) (#x01AA . #x01AB) (#x01B9 . #x01BA) (#x01BD . #x01BF) (#x01DC . #x01DD) (#x01EF . #x01F0) (#x0233 . #x0239) (#x023F . #x0240) (#x024F . #x0293) (#x0295 . #x02AF) (#x037B . #x037D) (#x03AC . #x03CE) (#x03D0 . #x03D1) (#x03D5 . #x03D7) (#x03EF . #x03F3) (#x03FB . #x03FC) (#x0430 . #x045F) (#x04CE . #x04CF) (#x0561 . #x0587) (#x1D00 . #x1D2B) (#x1D62 . #x1D77) (#x1D79 . #x1D9A) (#x1E95 . #x1E9D) (#x1EFF . #x1F07) (#x1F10 . #x1F15) (#x1F20 . #x1F27) (#x1F30 . #x1F37) (#x1F40 . #x1F45) (#x1F50 . #x1F57) (#x1F60 . #x1F67) (#x1F70 . #x1F7D) (#x1F80 . #x1F87) (#x1F90 . #x1F97) (#x1FA0 . #x1FA7) (#x1FB0 . #x1FB4) (#x1FB6 . #x1FB7) (#x1FC2 . #x1FC4) (#x1FC6 . #x1FC7) (#x1FD0 . #x1FD3) (#x1FD6 . #x1FD7) (#x1FE0 . #x1FE7) (#x1FF2 . #x1FF4) (#x1FF6 . #x1FF7) (#x210E . #x210F) (#x213C . #x213D) (#x2146 . #x2149) (#x2C30 . #x2C5E) (#x2C65 . #x2C66) (#x2C73 . #x2C74) (#x2C76 . #x2C7C) (#x2CE3 . #x2CE4) (#x2D00 . #x2D25) (#xA72F . #xA731) (#xA771 . #xA778) (#xFB00 . #xFB06) (#xFB13 . #xFB17) (#xFF41 . #xFF5A) (#x10428 . #x1044F) (#x1D41A . #x1D433) (#x1D44E . #x1D454) (#x1D456 . #x1D467) (#x1D482 . #x1D49B) (#x1D4B6 . #x1D4B9) (#x1D4BD . #x1D4C3) (#x1D4C5 . #x1D4CF) (#x1D4EA . #x1D503) (#x1D51E . #x1D537) (#x1D552 . #x1D56B) (#x1D586 . #x1D59F) (#x1D5BA . #x1D5D3) (#x1D5EE . #x1D607) (#x1D622 . #x1D63B) (#x1D656 . #x1D66F) (#x1D68A . #x1D6A5) (#x1D6C2 . #x1D6DA) (#x1D6DC . #x1D6E1) (#x1D6FC . #x1D714) (#x1D716 . #x1D71B) (#x1D736 . #x1D74E) (#x1D750 . #x1D755) (#x1D770 . #x1D788) (#x1D78A . #x1D78F) (#x1D7AA . #x1D7C2) (#x1D7C4 . #x1D7C9) #x00AA #x00B5 #x00BA #x0101 #x0103 #x0105 #x0107 #x0109 #x010B #x010D #x010F #x0111 #x0113 #x0115 #x0117 #x0119 #x011B #x011D #x011F #x0121 #x0123 #x0125 #x0127 #x0129 #x012B #x012D #x012F #x0131 #x0133 #x0135 #x013A #x013C #x013E #x0140 #x0142 #x0144 #x0146 #x014B #x014D #x014F #x0151 #x0153 #x0155 #x0157 #x0159 #x015B #x015D #x015F #x0161 #x0163 #x0165 #x0167 #x0169 #x016B #x016D #x016F #x0171 #x0173 #x0175 #x0177 #x017A #x017C #x0183 #x0185 #x0188 #x0192 #x0195 #x019E #x01A1 #x01A3 #x01A5 #x01A8 #x01AD #x01B0 #x01B4 #x01B6 #x01C6 #x01C9 #x01CC #x01CE #x01D0 #x01D2 #x01D4 #x01D6 #x01D8 #x01DA #x01DF #x01E1 #x01E3 #x01E5 #x01E7 #x01E9 #x01EB #x01ED #x01F3 #x01F5 #x01F9 #x01FB #x01FD #x01FF #x0201 #x0203 #x0205 #x0207 #x0209 #x020B #x020D #x020F #x0211 #x0213 #x0215 #x0217 #x0219 #x021B #x021D #x021F #x0221 #x0223 #x0225 #x0227 #x0229 #x022B #x022D #x022F #x0231 #x023C #x0242 #x0247 #x0249 #x024B #x024D #x0371 #x0373 #x0377 #x0390 #x03D9 #x03DB #x03DD #x03DF #x03E1 #x03E3 #x03E5 #x03E7 #x03E9 #x03EB #x03ED #x03F5 #x03F8 #x0461 #x0463 #x0465 #x0467 #x0469 #x046B #x046D #x046F #x0471 #x0473 #x0475 #x0477 #x0479 #x047B #x047D #x047F #x0481 #x048B #x048D #x048F #x0491 #x0493 #x0495 #x0497 #x0499 #x049B #x049D #x049F #x04A1 #x04A3 #x04A5 #x04A7 #x04A9 #x04AB #x04AD #x04AF #x04B1 #x04B3 #x04B5 #x04B7 #x04B9 #x04BB #x04BD #x04BF #x04C2 #x04C4 #x04C6 #x04C8 #x04CA #x04CC #x04D1 #x04D3 #x04D5 #x04D7 #x04D9 #x04DB #x04DD #x04DF #x04E1 #x04E3 #x04E5 #x04E7 #x04E9 #x04EB #x04ED #x04EF #x04F1 #x04F3 #x04F5 #x04F7 #x04F9 #x04FB #x04FD #x04FF #x0501 #x0503 #x0505 #x0507 #x0509 #x050B #x050D #x050F #x0511 #x0513 #x0515 #x0517 #x0519 #x051B #x051D #x051F #x0521 #x0523 #x1E01 #x1E03 #x1E05 #x1E07 #x1E09 #x1E0B #x1E0D #x1E0F #x1E11 #x1E13 #x1E15 #x1E17 #x1E19 #x1E1B #x1E1D #x1E1F #x1E21 #x1E23 #x1E25 #x1E27 #x1E29 #x1E2B #x1E2D #x1E2F #x1E31 #x1E33 #x1E35 #x1E37 #x1E39 #x1E3B #x1E3D #x1E3F #x1E41 #x1E43 #x1E45 #x1E47 #x1E49 #x1E4B #x1E4D #x1E4F #x1E51 #x1E53 #x1E55 #x1E57 #x1E59 #x1E5B #x1E5D #x1E5F #x1E61 #x1E63 #x1E65 #x1E67 #x1E69 #x1E6B #x1E6D #x1E6F #x1E71 #x1E73 #x1E75 #x1E77 #x1E79 #x1E7B #x1E7D #x1E7F #x1E81 #x1E83 #x1E85 #x1E87 #x1E89 #x1E8B #x1E8D #x1E8F #x1E91 #x1E93 #x1E9F #x1EA1 #x1EA3 #x1EA5 #x1EA7 #x1EA9 #x1EAB #x1EAD #x1EAF #x1EB1 #x1EB3 #x1EB5 #x1EB7 #x1EB9 #x1EBB #x1EBD #x1EBF #x1EC1 #x1EC3 #x1EC5 #x1EC7 #x1EC9 #x1ECB #x1ECD #x1ECF #x1ED1 #x1ED3 #x1ED5 #x1ED7 #x1ED9 #x1EDB #x1EDD #x1EDF #x1EE1 #x1EE3 #x1EE5 #x1EE7 #x1EE9 #x1EEB #x1EED #x1EEF #x1EF1 #x1EF3 #x1EF5 #x1EF7 #x1EF9 #x1EFB #x1EFD #x1FBE #x2071 #x207F #x210A #x2113 #x212F #x2134 #x2139 #x214E #x2184 #x2C61 #x2C68 #x2C6A #x2C6C #x2C71 #x2C81 #x2C83 #x2C85 #x2C87 #x2C89 #x2C8B #x2C8D #x2C8F #x2C91 #x2C93 #x2C95 #x2C97 #x2C99 #x2C9B #x2C9D #x2C9F #x2CA1 #x2CA3 #x2CA5 #x2CA7 #x2CA9 #x2CAB #x2CAD #x2CAF #x2CB1 #x2CB3 #x2CB5 #x2CB7 #x2CB9 #x2CBB #x2CBD #x2CBF #x2CC1 #x2CC3 #x2CC5 #x2CC7 #x2CC9 #x2CCB #x2CCD #x2CCF #x2CD1 #x2CD3 #x2CD5 #x2CD7 #x2CD9 #x2CDB #x2CDD #x2CDF #x2CE1 #xA641 #xA643 #xA645 #xA647 #xA649 #xA64B #xA64D #xA64F #xA651 #xA653 #xA655 #xA657 #xA659 #xA65B #xA65D #xA65F #xA663 #xA665 #xA667 #xA669 #xA66B #xA66D #xA681 #xA683 #xA685 #xA687 #xA689 #xA68B #xA68D #xA68F #xA691 #xA693 #xA695 #xA697 #xA723 #xA725 #xA727 #xA729 #xA72B #xA72D #xA733 #xA735 #xA737 #xA739 #xA73B #xA73D #xA73F #xA741 #xA743 #xA745 #xA747 #xA749 #xA74B #xA74D #xA74F #xA751 #xA753 #xA755 #xA757 #xA759 #xA75B #xA75D #xA75F #xA761 #xA763 #xA765 #xA767 #xA769 #xA76B #xA76D #xA76F #xA77A #xA77C #xA77F #xA781 #xA783 #xA785 #xA787 #xA78C #x1D4BB #x1D7CB)))
 
(define-char-set char-set:upper-case
  (decompress '#((#x0041 . #x005A) (#x00C0 . #x00D6) (#x00D8 . #x00DE) (#x0178 . #x0179) (#x0181 . #x0182) (#x0186 . #x0187) (#x0189 . #x018B) (#x018E . #x0191) (#x0193 . #x0194) (#x0196 . #x0198) (#x019C . #x019D) (#x019F . #x01A0) (#x01A6 . #x01A7) (#x01AE . #x01AF) (#x01B1 . #x01B3) (#x01B7 . #x01B8) (#x01F6 . #x01F8) (#x023A . #x023B) (#x023D . #x023E) (#x0243 . #x0246) (#x0388 . #x038A) (#x038E . #x038F) (#x0391 . #x03A1) (#x03A3 . #x03AB) (#x03D2 . #x03D4) (#x03F9 . #x03FA) (#x03FD . #x042F) (#x04C0 . #x04C1) (#x0531 . #x0556) (#x10A0 . #x10C5) (#x1F08 . #x1F0F) (#x1F18 . #x1F1D) (#x1F28 . #x1F2F) (#x1F38 . #x1F3F) (#x1F48 . #x1F4D) (#x1F68 . #x1F6F) (#x1FB8 . #x1FBB) (#x1FC8 . #x1FCB) (#x1FD8 . #x1FDB) (#x1FE8 . #x1FEC) (#x1FF8 . #x1FFB) (#x210B . #x210D) (#x2110 . #x2112) (#x2119 . #x211D) (#x212A . #x212D) (#x2130 . #x2133) (#x213E . #x213F) (#x2C00 . #x2C2E) (#x2C62 . #x2C64) (#x2C6D . #x2C6F) (#xA77D . #xA77E) (#xFF21 . #xFF3A) (#x10400 . #x10427) (#x1D400 . #x1D419) (#x1D434 . #x1D44D) (#x1D468 . #x1D481) (#x1D49E . #x1D49F) (#x1D4A5 . #x1D4A6) (#x1D4A9 . #x1D4AC) (#x1D4AE . #x1D4B5) (#x1D4D0 . #x1D4E9) (#x1D504 . #x1D505) (#x1D507 . #x1D50A) (#x1D50D . #x1D514) (#x1D516 . #x1D51C) (#x1D538 . #x1D539) (#x1D53B . #x1D53E) (#x1D540 . #x1D544) (#x1D54A . #x1D550) (#x1D56C . #x1D585) (#x1D5A0 . #x1D5B9) (#x1D5D4 . #x1D5ED) (#x1D608 . #x1D621) (#x1D63C . #x1D655) (#x1D670 . #x1D689) (#x1D6A8 . #x1D6C0) (#x1D6E2 . #x1D6FA) (#x1D71C . #x1D734) (#x1D756 . #x1D76E) (#x1D790 . #x1D7A8) #x0100 #x0102 #x0104 #x0106 #x0108 #x010A #x010C #x010E #x0110 #x0112 #x0114 #x0116 #x0118 #x011A #x011C #x011E #x0120 #x0122 #x0124 #x0126 #x0128 #x012A #x012C #x012E #x0130 #x0132 #x0134 #x0136 #x0139 #x013B #x013D #x013F #x0141 #x0143 #x0145 #x0147 #x014A #x014C #x014E #x0150 #x0152 #x0154 #x0156 #x0158 #x015A #x015C #x015E #x0160 #x0162 #x0164 #x0166 #x0168 #x016A #x016C #x016E #x0170 #x0172 #x0174 #x0176 #x017B #x017D #x0184 #x01A2 #x01A4 #x01A9 #x01AC #x01B5 #x01BC #x01C4 #x01C7 #x01CA #x01CD #x01CF #x01D1 #x01D3 #x01D5 #x01D7 #x01D9 #x01DB #x01DE #x01E0 #x01E2 #x01E4 #x01E6 #x01E8 #x01EA #x01EC #x01EE #x01F1 #x01F4 #x01FA #x01FC #x01FE #x0200 #x0202 #x0204 #x0206 #x0208 #x020A #x020C #x020E #x0210 #x0212 #x0214 #x0216 #x0218 #x021A #x021C #x021E #x0220 #x0222 #x0224 #x0226 #x0228 #x022A #x022C #x022E #x0230 #x0232 #x0241 #x0248 #x024A #x024C #x024E #x0370 #x0372 #x0376 #x0386 #x038C #x03CF #x03D8 #x03DA #x03DC #x03DE #x03E0 #x03E2 #x03E4 #x03E6 #x03E8 #x03EA #x03EC #x03EE #x03F4 #x03F7 #x0460 #x0462 #x0464 #x0466 #x0468 #x046A #x046C #x046E #x0470 #x0472 #x0474 #x0476 #x0478 #x047A #x047C #x047E #x0480 #x048A #x048C #x048E #x0490 #x0492 #x0494 #x0496 #x0498 #x049A #x049C #x049E #x04A0 #x04A2 #x04A4 #x04A6 #x04A8 #x04AA #x04AC #x04AE #x04B0 #x04B2 #x04B4 #x04B6 #x04B8 #x04BA #x04BC #x04BE #x04C3 #x04C5 #x04C7 #x04C9 #x04CB #x04CD #x04D0 #x04D2 #x04D4 #x04D6 #x04D8 #x04DA #x04DC #x04DE #x04E0 #x04E2 #x04E4 #x04E6 #x04E8 #x04EA #x04EC #x04EE #x04F0 #x04F2 #x04F4 #x04F6 #x04F8 #x04FA #x04FC #x04FE #x0500 #x0502 #x0504 #x0506 #x0508 #x050A #x050C #x050E #x0510 #x0512 #x0514 #x0516 #x0518 #x051A #x051C #x051E #x0520 #x0522 #x1E00 #x1E02 #x1E04 #x1E06 #x1E08 #x1E0A #x1E0C #x1E0E #x1E10 #x1E12 #x1E14 #x1E16 #x1E18 #x1E1A #x1E1C #x1E1E #x1E20 #x1E22 #x1E24 #x1E26 #x1E28 #x1E2A #x1E2C #x1E2E #x1E30 #x1E32 #x1E34 #x1E36 #x1E38 #x1E3A #x1E3C #x1E3E #x1E40 #x1E42 #x1E44 #x1E46 #x1E48 #x1E4A #x1E4C #x1E4E #x1E50 #x1E52 #x1E54 #x1E56 #x1E58 #x1E5A #x1E5C #x1E5E #x1E60 #x1E62 #x1E64 #x1E66 #x1E68 #x1E6A #x1E6C #x1E6E #x1E70 #x1E72 #x1E74 #x1E76 #x1E78 #x1E7A #x1E7C #x1E7E #x1E80 #x1E82 #x1E84 #x1E86 #x1E88 #x1E8A #x1E8C #x1E8E #x1E90 #x1E92 #x1E94 #x1E9E #x1EA0 #x1EA2 #x1EA4 #x1EA6 #x1EA8 #x1EAA #x1EAC #x1EAE #x1EB0 #x1EB2 #x1EB4 #x1EB6 #x1EB8 #x1EBA #x1EBC #x1EBE #x1EC0 #x1EC2 #x1EC4 #x1EC6 #x1EC8 #x1ECA #x1ECC #x1ECE #x1ED0 #x1ED2 #x1ED4 #x1ED6 #x1ED8 #x1EDA #x1EDC #x1EDE #x1EE0 #x1EE2 #x1EE4 #x1EE6 #x1EE8 #x1EEA #x1EEC #x1EEE #x1EF0 #x1EF2 #x1EF4 #x1EF6 #x1EF8 #x1EFA #x1EFC #x1EFE #x1F59 #x1F5B #x1F5D #x1F5F #x2102 #x2107 #x2115 #x2124 #x2126 #x2128 #x2145 #x2183 #x2C60 #x2C67 #x2C69 #x2C6B #x2C72 #x2C75 #x2C80 #x2C82 #x2C84 #x2C86 #x2C88 #x2C8A #x2C8C #x2C8E #x2C90 #x2C92 #x2C94 #x2C96 #x2C98 #x2C9A #x2C9C #x2C9E #x2CA0 #x2CA2 #x2CA4 #x2CA6 #x2CA8 #x2CAA #x2CAC #x2CAE #x2CB0 #x2CB2 #x2CB4 #x2CB6 #x2CB8 #x2CBA #x2CBC #x2CBE #x2CC0 #x2CC2 #x2CC4 #x2CC6 #x2CC8 #x2CCA #x2CCC #x2CCE #x2CD0 #x2CD2 #x2CD4 #x2CD6 #x2CD8 #x2CDA #x2CDC #x2CDE #x2CE0 #x2CE2 #xA640 #xA642 #xA644 #xA646 #xA648 #xA64A #xA64C #xA64E #xA650 #xA652 #xA654 #xA656 #xA658 #xA65A #xA65C #xA65E #xA662 #xA664 #xA666 #xA668 #xA66A #xA66C #xA680 #xA682 #xA684 #xA686 #xA688 #xA68A #xA68C #xA68E #xA690 #xA692 #xA694 #xA696 #xA722 #xA724 #xA726 #xA728 #xA72A #xA72C #xA72E #xA732 #xA734 #xA736 #xA738 #xA73A #xA73C #xA73E #xA740 #xA742 #xA744 #xA746 #xA748 #xA74A #xA74C #xA74E #xA750 #xA752 #xA754 #xA756 #xA758 #xA75A #xA75C #xA75E #xA760 #xA762 #xA764 #xA766 #xA768 #xA76A #xA76C #xA76E #xA779 #xA77B #xA780 #xA782 #xA784 #xA786 #xA78B #x1D49C #x1D4A2 #x1D546 #x1D7CA)))

(define-char-set char-set:title-case
  (decompress '#((#x1F88 . #x1F8F) (#x1F98 . #x1F9F) (#x1FA8 . #x1FAF) #x01C5 #x01C8 #x01CB #x01F2 #x1FBC #x1FCC #x1FFC)))

(define-char-set char-set:letter
  (let* ((s (t-union (char-set:s char-set:lower-case)
                     (char-set:s char-set:upper-case)))
         (s (t-union s (char-set:s char-set:title-case)))
         (s (decompress '#((#x02B0 . #x02C1) (#x02C6 . #x02D1) (#x02E0 . #x02E4) (#x06E5 . #x06E6) (#x07F4 . #x07F5) (#x1C78 . #x1C7D) (#x1D2C . #x1D61) (#x1D9B . #x1DBF) (#x2090 . #x2094) (#x3031 . #x3035) (#x309D . #x309E) (#x30FC . #x30FE) (#xA717 . #xA71F) (#xFF9E . #xFF9F) #x02EC #x02EE #x0374 #x037A #x0559 #x0640 #x07FA #x0971 #x0E46 #x0EC6 #x10FC #x17D7 #x1843 #x1D78 #x2C7D #x2D6F #x2E2F #x3005 #x303B #xA015 #xA60C #xA67F #xA770 #xA788 #xFF70) s))) ; Lm
    (decompress '#((#x01C0 . #x01C3) (#x05D0 . #x05EA) (#x05F0 . #x05F2) (#x0621 . #x063F) (#x0641 . #x064A) (#x066E . #x066F) (#x0671 . #x06D3) (#x06EE . #x06EF) (#x06FA . #x06FC) (#x0712 . #x072F) (#x074D . #x07A5) (#x07CA . #x07EA) (#x0904 . #x0939) (#x0958 . #x0961) (#x097B . #x097F) (#x0985 . #x098C) (#x098F . #x0990) (#x0993 . #x09A8) (#x09AA . #x09B0) (#x09B6 . #x09B9) (#x09DC . #x09DD) (#x09DF . #x09E1) (#x09F0 . #x09F1) (#x0A05 . #x0A0A) (#x0A0F . #x0A10) (#x0A13 . #x0A28) (#x0A2A . #x0A30) (#x0A32 . #x0A33) (#x0A35 . #x0A36) (#x0A38 . #x0A39) (#x0A59 . #x0A5C) (#x0A72 . #x0A74) (#x0A85 . #x0A8D) (#x0A8F . #x0A91) (#x0A93 . #x0AA8) (#x0AAA . #x0AB0) (#x0AB2 . #x0AB3) (#x0AB5 . #x0AB9) (#x0AE0 . #x0AE1) (#x0B05 . #x0B0C) (#x0B0F . #x0B10) (#x0B13 . #x0B28) (#x0B2A . #x0B30) (#x0B32 . #x0B33) (#x0B35 . #x0B39) (#x0B5C . #x0B5D) (#x0B5F . #x0B61) (#x0B85 . #x0B8A) (#x0B8E . #x0B90) (#x0B92 . #x0B95) (#x0B99 . #x0B9A) (#x0B9E . #x0B9F) (#x0BA3 . #x0BA4) (#x0BA8 . #x0BAA) (#x0BAE . #x0BB9) (#x0C05 . #x0C0C) (#x0C0E . #x0C10) (#x0C12 . #x0C28) (#x0C2A . #x0C33) (#x0C35 . #x0C39) (#x0C58 . #x0C59) (#x0C60 . #x0C61) (#x0C85 . #x0C8C) (#x0C8E . #x0C90) (#x0C92 . #x0CA8) (#x0CAA . #x0CB3) (#x0CB5 . #x0CB9) (#x0CE0 . #x0CE1) (#x0D05 . #x0D0C) (#x0D0E . #x0D10) (#x0D12 . #x0D28) (#x0D2A . #x0D39) (#x0D60 . #x0D61) (#x0D7A . #x0D7F) (#x0D85 . #x0D96) (#x0D9A . #x0DB1) (#x0DB3 . #x0DBB) (#x0DC0 . #x0DC6) (#x0E01 . #x0E30) (#x0E32 . #x0E33) (#x0E40 . #x0E45) (#x0E81 . #x0E82) (#x0E87 . #x0E88) (#x0E94 . #x0E97) (#x0E99 . #x0E9F) (#x0EA1 . #x0EA3) (#x0EAA . #x0EAB) (#x0EAD . #x0EB0) (#x0EB2 . #x0EB3) (#x0EC0 . #x0EC4) (#x0EDC . #x0EDD) (#x0F40 . #x0F47) (#x0F49 . #x0F6C) (#x0F88 . #x0F8B) (#x1000 . #x102A) (#x1050 . #x1055) (#x105A . #x105D) (#x1065 . #x1066) (#x106E . #x1070) (#x1075 . #x1081) (#x10D0 . #x10FA) (#x1100 . #x1159) (#x115F . #x11A2) (#x11A8 . #x11F9) (#x1200 . #x1248) (#x124A . #x124D) (#x1250 . #x1256) (#x125A . #x125D) (#x1260 . #x1288) (#x128A . #x128D) (#x1290 . #x12B0) (#x12B2 . #x12B5) (#x12B8 . #x12BE) (#x12C2 . #x12C5) (#x12C8 . #x12D6) (#x12D8 . #x1310) (#x1312 . #x1315) (#x1318 . #x135A) (#x1380 . #x138F) (#x13A0 . #x13F4) (#x1401 . #x166C) (#x166F . #x1676) (#x1681 . #x169A) (#x16A0 . #x16EA) (#x1700 . #x170C) (#x170E . #x1711) (#x1720 . #x1731) (#x1740 . #x1751) (#x1760 . #x176C) (#x176E . #x1770) (#x1780 . #x17B3) (#x1820 . #x1842) (#x1844 . #x1877) (#x1880 . #x18A8) (#x1900 . #x191C) (#x1950 . #x196D) (#x1970 . #x1974) (#x1980 . #x19A9) (#x19C1 . #x19C7) (#x1A00 . #x1A16) (#x1B05 . #x1B33) (#x1B45 . #x1B4B) (#x1B83 . #x1BA0) (#x1BAE . #x1BAF) (#x1C00 . #x1C23) (#x1C4D . #x1C4F) (#x1C5A . #x1C77) (#x2135 . #x2138) (#x2D30 . #x2D65) (#x2D80 . #x2D96) (#x2DA0 . #x2DA6) (#x2DA8 . #x2DAE) (#x2DB0 . #x2DB6) (#x2DB8 . #x2DBE) (#x2DC0 . #x2DC6) (#x2DC8 . #x2DCE) (#x2DD0 . #x2DD6) (#x2DD8 . #x2DDE) (#x3041 . #x3096) (#x30A1 . #x30FA) (#x3105 . #x312D) (#x3131 . #x318E) (#x31A0 . #x31B7) (#x31F0 . #x31FF) (#xA000 . #xA014) (#xA016 . #xA48C) (#xA500 . #xA60B) (#xA610 . #xA61F) (#xA62A . #xA62B) (#xA7FB . #xA801) (#xA803 . #xA805) (#xA807 . #xA80A) (#xA80C . #xA822) (#xA840 . #xA873) (#xA882 . #xA8B3) (#xA90A . #xA925) (#xA930 . #xA946) (#xAA00 . #xAA28) (#xAA40 . #xAA42) (#xAA44 . #xAA4B) (#xF900 . #xFA2D) (#xFA30 . #xFA6A) (#xFA70 . #xFAD9) (#xFB1F . #xFB28) (#xFB2A . #xFB36) (#xFB38 . #xFB3C) (#xFB40 . #xFB41) (#xFB43 . #xFB44) (#xFB46 . #xFBB1) (#xFBD3 . #xFD3D) (#xFD50 . #xFD8F) (#xFD92 . #xFDC7) (#xFDF0 . #xFDFB) (#xFE70 . #xFE74) (#xFE76 . #xFEFC) (#xFF66 . #xFF6F) (#xFF71 . #xFF9D) (#xFFA0 . #xFFBE) (#xFFC2 . #xFFC7) (#xFFCA . #xFFCF) (#xFFD2 . #xFFD7) (#xFFDA . #xFFDC) (#x10000 . #x1000B) (#x1000D . #x10026) (#x10028 . #x1003A) (#x1003C . #x1003D) (#x1003F . #x1004D) (#x10050 . #x1005D) (#x10080 . #x100FA) (#x10280 . #x1029C) (#x102A0 . #x102D0) (#x10300 . #x1031E) (#x10330 . #x10340) (#x10342 . #x10349) (#x10380 . #x1039D) (#x103A0 . #x103C3) (#x103C8 . #x103CF) (#x10450 . #x1049D) (#x10800 . #x10805) (#x1080A . #x10835) (#x10837 . #x10838) (#x10900 . #x10915) (#x10920 . #x10939) (#x10A10 . #x10A13) (#x10A15 . #x10A17) (#x10A19 . #x10A33) (#x12000 . #x1236E) (#x2F800 . #x2FA1D) #x01BB #x0294 #x06D5 #x06FF #x0710 #x07B1 #x093D #x0950 #x0972 #x09B2 #x09BD #x09CE #x0A5E #x0ABD #x0AD0 #x0B3D #x0B71 #x0B83 #x0B9C #x0BD0 #x0C3D #x0CBD #x0CDE #x0D3D #x0DBD #x0E84 #x0E8A #x0E8D #x0EA5 #x0EA7 #x0EBD #x0F00 #x103F #x1061 #x108E #x1258 #x12C0 #x17DC #x18AA #x3006 #x303C #x309F #x30FF (#x3400 . #x4DB5) (#x4E00 . #x9FC3) #xA66E (#xAC00 . #xD7A3) #xFB1D #xFB3E #x10808 #x1083C #x1083F #x10A00 (#x20000 . #x2A6D6)) s)))

(define-char-set char-set:digit
  (decompress '#((#x0030 . #x0039) (#x0660 . #x0669) (#x06F0 . #x06F9) (#x07C0 . #x07C9) (#x0966 . #x096F) (#x09E6 . #x09EF) (#x0A66 . #x0A6F) (#x0AE6 . #x0AEF) (#x0B66 . #x0B6F) (#x0BE6 . #x0BEF) (#x0C66 . #x0C6F) (#x0CE6 . #x0CEF) (#x0D66 . #x0D6F) (#x0E50 . #x0E59) (#x0ED0 . #x0ED9) (#x0F20 . #x0F29) (#x1040 . #x1049) (#x1090 . #x1099) (#x17E0 . #x17E9) (#x1810 . #x1819) (#x1946 . #x194F) (#x19D0 . #x19D9) (#x1B50 . #x1B59) (#x1BB0 . #x1BB9) (#x1C40 . #x1C49) (#x1C50 . #x1C59) (#xA620 . #xA629) (#xA8D0 . #xA8D9) (#xA900 . #xA909) (#xAA50 . #xAA59) (#xFF10 . #xFF19) (#x104A0 . #x104A9) (#x1D7CE . #x1D7FF))))

(define char-set:hex-digit (string->char-set "0123456789abcdefABCDEF"))

(define char-set:letter+digit
  (make-char-set (delay (##sys#slot (char-set-union char-set:letter char-set:digit) 1))))

(define-char-set char-set:punctuation
  (decompress '#((#x0021 . #x0023) (#x0025 . #x002A) (#x002C . #x002F) (#x003A . #x003B) (#x003F . #x0040) (#x005B . #x005D) (#x055A . #x055F) (#x0589 . #x058A) (#x05F3 . #x05F4) (#x0609 . #x060A) (#x060C . #x060D) (#x061E . #x061F) (#x066A . #x066D) (#x0700 . #x070D) (#x07F7 . #x07F9) (#x0964 . #x0965) (#x0E5A . #x0E5B) (#x0F04 . #x0F12) (#x0F3A . #x0F3D) (#x0FD0 . #x0FD4) (#x104A . #x104F) (#x1361 . #x1368) (#x166D . #x166E) (#x169B . #x169C) (#x16EB . #x16ED) (#x1735 . #x1736) (#x17D4 . #x17D6) (#x17D8 . #x17DA) (#x1800 . #x180A) (#x1944 . #x1945) (#x19DE . #x19DF) (#x1A1E . #x1A1F) (#x1B5A . #x1B60) (#x1C3B . #x1C3F) (#x1C7E . #x1C7F) (#x2010 . #x2027) (#x2030 . #x2043) (#x2045 . #x2051) (#x2053 . #x205E) (#x207D . #x207E) (#x208D . #x208E) (#x2329 . #x232A) (#x2768 . #x2775) (#x27C5 . #x27C6) (#x27E6 . #x27EF) (#x2983 . #x2998) (#x29D8 . #x29DB) (#x29FC . #x29FD) (#x2CF9 . #x2CFC) (#x2CFE . #x2CFF) (#x2E00 . #x2E2E) (#x3001 . #x3003) (#x3008 . #x3011) (#x3014 . #x301F) (#xA60D . #xA60F) (#xA874 . #xA877) (#xA8CE . #xA8CF) (#xA92E . #xA92F) (#xAA5C . #xAA5F) (#xFD3E . #xFD3F) (#xFE10 . #xFE19) (#xFE30 . #xFE52) (#xFE54 . #xFE61) (#xFE6A . #xFE6B) (#xFF01 . #xFF03) (#xFF05 . #xFF0A) (#xFF0C . #xFF0F) (#xFF1A . #xFF1B) (#xFF1F . #xFF20) (#xFF3B . #xFF3D) (#xFF5F . #xFF65) (#x10100 . #x10101) (#x10A50 . #x10A58) (#x12470 . #x12473) #x005F #x007B #x007D #x00A1 #x00AB #x00B7 #x00BB #x00BF #x037E #x0387 #x05BE #x05C0 #x05C3 #x05C6 #x061B #x06D4 #x0970 #x0DF4 #x0E4F #x0F85 #x10FB #x2E30 #x3030 #x303D #x30A0 #x30FB #xA673 #xA67E #xA95F #xFE63 #xFE68 #xFF3F #xFF5B #xFF5D #x1039F #x103D0 #x1091F #x1093F)))

(define-char-set char-set:symbol
  (decompress '#((#x003C . #x003E) (#x00A2 . #x00A9) (#x00AE . #x00B1) (#x02C2 . #x02C5) (#x02D2 . #x02DF) (#x02E5 . #x02EB) (#x02EF . #x02FF) (#x0384 . #x0385) (#x0606 . #x0608) (#x060E . #x060F) (#x06FD . #x06FE) (#x09F2 . #x09F3) (#x0BF3 . #x0BFA) (#x0CF1 . #x0CF2) (#x0F01 . #x0F03) (#x0F13 . #x0F17) (#x0F1A . #x0F1F) (#x0FBE . #x0FC5) (#x0FC7 . #x0FCC) (#x0FCE . #x0FCF) (#x109E . #x109F) (#x1390 . #x1399) (#x19E0 . #x19FF) (#x1B61 . #x1B6A) (#x1B74 . #x1B7C) (#x1FBF . #x1FC1) (#x1FCD . #x1FCF) (#x1FDD . #x1FDF) (#x1FED . #x1FEF) (#x1FFD . #x1FFE) (#x207A . #x207C) (#x208A . #x208C) (#x20A0 . #x20B5) (#x2100 . #x2101) (#x2103 . #x2106) (#x2108 . #x2109) (#x2116 . #x2118) (#x211E . #x2123) (#x213A . #x213B) (#x2140 . #x2144) (#x214A . #x214D) (#x2190 . #x2328) (#x232B . #x23E7) (#x2400 . #x2426) (#x2440 . #x244A) (#x249C . #x24E9) (#x2500 . #x269D) (#x26A0 . #x26BC) (#x26C0 . #x26C3) (#x2701 . #x2704) (#x2706 . #x2709) (#x270C . #x2727) (#x2729 . #x274B) (#x274F . #x2752) (#x2758 . #x275E) (#x2761 . #x2767) (#x2798 . #x27AF) (#x27B1 . #x27BE) (#x27C0 . #x27C4) (#x27C7 . #x27CA) (#x27D0 . #x27E5) (#x27F0 . #x2982) (#x2999 . #x29D7) (#x29DC . #x29FB) (#x29FE . #x2B4C) (#x2B50 . #x2B54) (#x2CE5 . #x2CEA) (#x2E80 . #x2E99) (#x2E9B . #x2EF3) (#x2F00 . #x2FD5) (#x2FF0 . #x2FFB) (#x3012 . #x3013) (#x3036 . #x3037) (#x303E . #x303F) (#x309B . #x309C) (#x3190 . #x3191) (#x3196 . #x319F) (#x31C0 . #x31E3) (#x3200 . #x321E) (#x322A . #x3243) (#x3260 . #x327F) (#x328A . #x32B0) (#x32C0 . #x32FE) (#x3300 . #x33FF) (#x4DC0 . #x4DFF) (#xA490 . #xA4C6) (#xA700 . #xA716) (#xA720 . #xA721) (#xA789 . #xA78A) (#xA828 . #xA82B) (#xFDFC . #xFDFD) (#xFE64 . #xFE66) (#xFF1C . #xFF1E) (#xFFE0 . #xFFE6) (#xFFE8 . #xFFEE) (#xFFFC . #xFFFD) (#x10137 . #x1013F) (#x10179 . #x10189) (#x10190 . #x1019B) (#x101D0 . #x101FC) (#x1D000 . #x1D0F5) (#x1D100 . #x1D126) (#x1D129 . #x1D164) (#x1D16A . #x1D16C) (#x1D183 . #x1D184) (#x1D18C . #x1D1A9) (#x1D1AE . #x1D1DD) (#x1D200 . #x1D241) (#x1D300 . #x1D356) (#x1F000 . #x1F02B) (#x1F030 . #x1F093) #x0024 #x002B #x005E #x0060 #x007C #x007E #x00AC #x00B4 #x00B6 #x00B8 #x00D7 #x00F7 #x02ED #x0375 #x03F6 #x0482 #x060B #x06E9 #x07F6 #x09FA #x0AF1 #x0B70 #x0C7F #x0D79 #x0E3F #x0F34 #x0F36 #x0F38 #x1360 #x17DB #x1940 #x1FBD #x2044 #x2052 #x2114 #x2125 #x2127 #x2129 #x212E #x214F #x274D #x2756 #x2794 #x27CC #x3004 #x3020 #x3250 #xFB29 #xFE62 #xFE69 #xFF04 #xFF0B #xFF3E #xFF40 #xFF5C #xFF5E #x10102 #x1D245 #x1D6C1 #x1D6DB #x1D6FB #x1D715 #x1D735 #x1D74F #x1D76F #x1D789 #x1D7A9 #x1D7C3)))

(define-char-set char-set:graphic
  (let* ((s (decompress '#((#x0030 . #x0039) (#x00B2 . #x00B3) (#x00BC . #x00BE) (#x0660 . #x0669) (#x06F0 . #x06F9) (#x07C0 . #x07C9) (#x0966 . #x096F) (#x09E6 . #x09EF) (#x09F4 . #x09F9) (#x0A66 . #x0A6F) (#x0AE6 . #x0AEF) (#x0B66 . #x0B6F) (#x0BE6 . #x0BF2) (#x0C66 . #x0C6F) (#x0C78 . #x0C7E) (#x0CE6 . #x0CEF) (#x0D66 . #x0D75) (#x0E50 . #x0E59) (#x0ED0 . #x0ED9) (#x0F20 . #x0F33) (#x1040 . #x1049) (#x1090 . #x1099) (#x1369 . #x137C) (#x16EE . #x16F0) (#x17E0 . #x17E9) (#x17F0 . #x17F9) (#x1810 . #x1819) (#x1946 . #x194F) (#x19D0 . #x19D9) (#x1B50 . #x1B59) (#x1BB0 . #x1BB9) (#x1C40 . #x1C49) (#x1C50 . #x1C59) (#x2074 . #x2079) (#x2080 . #x2089) (#x2153 . #x2182) (#x2185 . #x2188) (#x2460 . #x249B) (#x24EA . #x24FF) (#x2776 . #x2793) (#x3021 . #x3029) (#x3038 . #x303A) (#x3192 . #x3195) (#x3220 . #x3229) (#x3251 . #x325F) (#x3280 . #x3289) (#x32B1 . #x32BF) (#xA620 . #xA629) (#xA8D0 . #xA8D9) (#xA900 . #xA909) (#xAA50 . #xAA59) (#xFF10 . #xFF19) (#x10107 . #x10133) (#x10140 . #x10178) (#x10320 . #x10323) (#x103D1 . #x103D5) (#x104A0 . #x104A9) (#x10916 . #x10919) (#x10A40 . #x10A47) (#x12400 . #x12462) (#x1D360 . #x1D371) (#x1D7CE . #x1D7FF) #x00B9 #x2070 #x2CFD #x3007 #x1018A #x10341 #x1034A))) ; N*
         (s (decompress '#((#x0300 . #x036F) (#x0483 . #x0489) (#x0591 . #x05BD) (#x05C1 . #x05C2) (#x05C4 . #x05C5) (#x0610 . #x061A) (#x064B . #x065E) (#x06D6 . #x06DC) (#x06DE . #x06E4) (#x06E7 . #x06E8) (#x06EA . #x06ED) (#x0730 . #x074A) (#x07A6 . #x07B0) (#x07EB . #x07F3) (#x0901 . #x0903) (#x093E . #x094D) (#x0951 . #x0954) (#x0962 . #x0963) (#x0981 . #x0983) (#x09BE . #x09C4) (#x09C7 . #x09C8) (#x09CB . #x09CD) (#x09E2 . #x09E3) (#x0A01 . #x0A03) (#x0A3E . #x0A42) (#x0A47 . #x0A48) (#x0A4B . #x0A4D) (#x0A70 . #x0A71) (#x0A81 . #x0A83) (#x0ABE . #x0AC5) (#x0AC7 . #x0AC9) (#x0ACB . #x0ACD) (#x0AE2 . #x0AE3) (#x0B01 . #x0B03) (#x0B3E . #x0B44) (#x0B47 . #x0B48) (#x0B4B . #x0B4D) (#x0B56 . #x0B57) (#x0B62 . #x0B63) (#x0BBE . #x0BC2) (#x0BC6 . #x0BC8) (#x0BCA . #x0BCD) (#x0C01 . #x0C03) (#x0C3E . #x0C44) (#x0C46 . #x0C48) (#x0C4A . #x0C4D) (#x0C55 . #x0C56) (#x0C62 . #x0C63) (#x0C82 . #x0C83) (#x0CBE . #x0CC4) (#x0CC6 . #x0CC8) (#x0CCA . #x0CCD) (#x0CD5 . #x0CD6) (#x0CE2 . #x0CE3) (#x0D02 . #x0D03) (#x0D3E . #x0D44) (#x0D46 . #x0D48) (#x0D4A . #x0D4D) (#x0D62 . #x0D63) (#x0D82 . #x0D83) (#x0DCF . #x0DD4) (#x0DD8 . #x0DDF) (#x0DF2 . #x0DF3) (#x0E34 . #x0E3A) (#x0E47 . #x0E4E) (#x0EB4 . #x0EB9) (#x0EBB . #x0EBC) (#x0EC8 . #x0ECD) (#x0F18 . #x0F19) (#x0F3E . #x0F3F) (#x0F71 . #x0F84) (#x0F86 . #x0F87) (#x0F90 . #x0F97) (#x0F99 . #x0FBC) (#x102B . #x103E) (#x1056 . #x1059) (#x105E . #x1060) (#x1062 . #x1064) (#x1067 . #x106D) (#x1071 . #x1074) (#x1082 . #x108D) (#x1712 . #x1714) (#x1732 . #x1734) (#x1752 . #x1753) (#x1772 . #x1773) (#x17B6 . #x17D3) (#x180B . #x180D) (#x1920 . #x192B) (#x1930 . #x193B) (#x19B0 . #x19C0) (#x19C8 . #x19C9) (#x1A17 . #x1A1B) (#x1B00 . #x1B04) (#x1B34 . #x1B44) (#x1B6B . #x1B73) (#x1B80 . #x1B82) (#x1BA1 . #x1BAA) (#x1C24 . #x1C37) (#x1DC0 . #x1DE6) (#x1DFE . #x1DFF) (#x20D0 . #x20F0) (#x2DE0 . #x2DFF) (#x302A . #x302F) (#x3099 . #x309A) (#xA66F . #xA672) (#xA67C . #xA67D) (#xA823 . #xA827) (#xA880 . #xA881) (#xA8B4 . #xA8C4) (#xA926 . #xA92D) (#xA947 . #xA953) (#xAA29 . #xAA36) (#xAA4C . #xAA4D) (#xFE00 . #xFE0F) (#xFE20 . #xFE26) (#x10A01 . #x10A03) (#x10A05 . #x10A06) (#x10A0C . #x10A0F) (#x10A38 . #x10A3A) (#x1D165 . #x1D169) (#x1D16D . #x1D172) (#x1D17B . #x1D182) (#x1D185 . #x1D18B) (#x1D1AA . #x1D1AD) (#x1D242 . #x1D244) (#xE0100 . #xE01EF) #x05BF #x05C7 #x0670 #x0711 #x093C #x09BC #x09D7 #x0A3C #x0A51 #x0A75 #x0ABC #x0B3C #x0B82 #x0BD7 #x0CBC #x0D57 #x0DCA #x0DD6 #x0E31 #x0EB1 #x0F35 #x0F37 #x0F39 #x0FC6 #x108F #x135F #x17DD #x18A9 #xA802 #xA806 #xA80B #xAA43 #xFB1E #x101FD #x10A3F) s))) ; M*
    (t-union (char-set:s char-set:punctuation)
             (t-union (char-set:s char-set:letter)
                      (t-union (char-set:s char-set:symbol)
                               s)))))

(define-char-set char-set:whitespace
  (decompress '#((#x09 . #x0d) (#x2000 . #x200a) 
                                       #x20 ; SPACE
                                       #xA0 #x1680 #x180e #x2028 #x205f
                                       #x3000)))

(define-char-set char-set:printing
  (t-union (char-set:s char-set:whitespace)
           (char-set:s char-set:graphic)))

(define-char-set char-set:blank
  (decompress '#((#x2000 . #x200a) 9 #x0020 #x00A0 #x1680 #x180E #x202F #x205F #x3000)))

(define-char-set char-set:iso-control
  (decompress '#((#x7F . #xbf) (0 . 31))))

(define char-set:ascii (ucs-range->char-set 0 128))

) ; module

;;; Copyright notice
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Copyright (c) 1988-1995 Massachusetts Institute of Technology
;;; 
;;; This material was developed by the Scheme project at the Massachusetts
;;; Institute of Technology, Department of Electrical Engineering and
;;; Computer Science.  Permission to copy and modify this software, to
;;; redistribute either the original software or a modified version, and
;;; to use this software for any purpose is granted, subject to the
;;; following restrictions and understandings.
;;; 
;;; 1. Any copy made of this software must include this copyright notice
;;; in full.
;;; 
;;; 2. Users of this software agree to make their best efforts (a) to
;;; return to the MIT Scheme project any improvements or extensions that
;;; they make, so that these may be included in future releases; and (b)
;;; to inform MIT of noteworthy uses of this software.
;;; 
;;; 3. All materials developed as a consequence of the use of this
;;; software shall duly acknowledge such use, in accordance with the usual
;;; standards of acknowledging credit in academic research.
;;; 
;;; 4. MIT has made no warrantee or representation that the operation of
;;; this software will be error-free, and MIT is under no obligation to
;;; provide any services, by way of maintenance, update, or otherwise.
;;; 
;;; 5. In conjunction with products arising from the use of this material,
;;; there shall be no use of the name of the Massachusetts Institute of
;;; Technology nor of any adaptation thereof in any advertising,
;;; promotional, or sales literature without prior written consent from
;;; MIT in each case.
